with Golden, ScalGolden, Pooling;
use Golden;

pragma Elaborate_All (Golden,ScalGolden,Pooling);

generic

package Taylors1.Skews is --- skew-product maps via Taylor series

  pragma Elaborate_Body;

  subtype Index4 is Integer range 1 .. 4;
  type SMat2 is array (Index4) of Scalar;
  type TMat2 is array (Index4) of Taylor1;

  procedure Copy(F: in TMat2; G: in out TMat2);
  procedure Add(F: in TMat2; G: in out TMat2);
  procedure ShowEigenVal(N: in String; M: in SMat2; Det1: in Boolean := False);
  procedure Det(F: in TMat2; P0: in out Taylor1);
  procedure DDet(F,DF: in TMat2; P0: in out Taylor1);
  procedure Prod(M1,M2: in SMat2; M3: in out SMat2);
  procedure Prod(M: in SMat2; F: in TMat2; G: in out TMat2);
  procedure Prod(F: in TMat2; M: in SMat2; G: in out TMat2);
  procedure Prod(P: in Taylor1; F: in TMat2; G: in out TMat2);
  procedure Prod(F,G: in TMat2; H: in out TMat2);
  procedure Translate(F: in TMat2; S: in Scalar; G: in out TMat2; R: in Radius);
  procedure Conjugate(F: in out TMat2);
  procedure Fix_Reversible(F: in out TMat2);
  procedure Factors(M: in TMat2; UFac,VFac: in out Scalar);

  Allow_Parallel: Boolean := True;  --- applies to Prod and Translate

  package SG is new ScalGolden (Scalar => Scalar);
  use SG;

  procedure CP(F: in TMat2; G: in out TMat2) renames Copy;
  package TP is new Pooling (Item => TMat2);
  subtype Pooled_TMat2 is TP.Controlled_LD_Pointer;

  type Skew is new Pooled_TMat2 with
    --- L: LD_Pointer := null;   -- L.all.Data is the TMat2
    record
      Det1: Boolean; --- determinant 1
      Alpha: GNum;   --- circle rotation number
    end record;

  --- main representation
  JT: constant Index4 := 1;  -- T=(A+D)/2
  JU: constant Index4 := 2;  -- U=B
  JV: constant Index4 := 3;  -- V=C
  JS: constant Index4 := 4;  -- S=(A-D)/2

  --- used for products
  JA: Index4 renames JT; -- A=T+S
  JB: Index4 renames JU; -- B=U
  JC: Index4 renames JV; -- C=V
  JD: Index4 renames JS; -- D=T-S

  function CName(J: Index4) return Character;
  function IsZero(F: Skew) return Boolean;
  procedure SetZero(R: in Radius; F: in out Skew);       --- Alpha=0
  procedure Copy(F1: in Skew; F2: in out Skew);
  procedure Swap(F1,F2: in out Skew);
  procedure Neg(F: in out Skew);
  procedure Neg(F1: in Skew; F2: in out Skew);
  procedure Add(F1: in Skew; F2: in out Skew);
  procedure Sum(F1,F2: in Skew; F3: in out Skew);
  procedure Sub(F1: in Skew; F2: in out Skew);
  procedure Diff(F1,F2: in Skew; F3: in out Skew);
  procedure Val(F: in Skew; S: in Scalar; FS: in out SMat2);
  procedure Component(F: in Skew; I: in Power; J: in Index4; S: in out Scalar);
  procedure TraceAt0(F: in Skew; S: in out Scalar);
  procedure SameRho(F: in out Skew);
  function WarnRho(N: String; F: Skew) return Radius;
  function Rho(F: Skew) return Radius;
  procedure SetRho(R: in Radius; F: in out Skew; Check: in Boolean := True);
  procedure Norm(R: in Radius; F: in Skew; S: in out Scalar);
  procedure Norm(F: in Skew; S: in out Scalar);
  function MaxNorm(R: Radius; F: Skew) return Radius;
  function MaxNorm(F: Skew) return Radius;
  function MaxDist(R: Radius; F1,F2: Skew) return Radius;
  function MaxDist(F1,F2: Skew) return Radius;

  procedure Inv(F: in out Skew);
  procedure Inv(F1: in Skew; F2: in out Skew);
  procedure Prod(S: in Scalar; F1: in Skew; F2: in out Skew);
  procedure Mult(S: in Scalar; F: in out Skew);
  procedure Prod(P: in Taylor1; F1: in Skew; F2: in out Skew);
  procedure Mult(P: in Taylor1; F: in out Skew);
  procedure AddProd(P: in Taylor1; F1: in Skew; F2: in out Skew);
  procedure Prod(F1,F2: in Skew; F3: in out Skew; R: in Radius);
  procedure DProd(F,G,DF,DG: in Skew; DH: in out Skew; R: in Radius);
  procedure Reflect(F: in out Skew; Offset: in Parity);
  procedure Scale1(F: in out Skew);
  procedure Scale3(F: in out Skew);
  procedure Translate(F: in Skew; S: in Scalar; G: in out Skew; R: in Radius); --- Uses x+S
  function DetMinus1(F: Skew; R: Radius) return Radius;
  procedure Show_Determinant(N: in String; F: in Skew);
  procedure Fix_Rho(F: in out Skew);
  procedure Fix_Reversible(F: in out Skew);
  procedure Show_Reversible(N: in String; F: in Skew);

  procedure AddBall(R: in Radius; S: in Scalar; F: in out Skew);
  procedure NumTrunc(F: in out Skew; Deg: in Power := PDeg; Cut: in Radius := Zero);
  procedure NumInterpolate(F0: in Skew; F1: in out Skew; Fac: in Flt);
  procedure NumIProd(F1,F2: in Skew; S: in out Scalar);
  function GuessRho(F: Skew) return Radius;

  function Strng(N: GNum) return String;
  procedure ShowType1(N: in String; F: in Skew);
  procedure ShowType2(N: in String; F1,F2: in Skew);
  procedure ShowNorms(N: in String; F: in Skew; ShowType: in Boolean := True);
  procedure Show1(N: in String; F: in Skew);
  procedure Show2(N: in String; F1,F2: in Skew);
  procedure Put(F: in File_Type; G: in Skew; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; G: in out Skew; Decimal: in Boolean := False);
  procedure Write(FileName: in String; G: in Skew; Decimal: in Boolean := False);
  procedure Read(FileName: in String; G: in out Skew; Decimal: in Boolean := False);

  procedure PutPlot(F: in File_Type; P: in Taylor1; XL,XR: in Flt);
  procedure WritePlot(FileName: in String; P: in Taylor1; XL,XR: in Flt);
  procedure WritePlots(Prefix: in String; F: in Skew; XL,XR: in Flt);

  procedure FZero(F: in out Skew; R: in Radius);
  procedure RevCosAM(E: in Scalar; F: in out Skew; R: in Radius);
  procedure Equalize(F,G: in out Skew; UFac,VFac: in out Scalar);
  procedure Normalize(F1: in Skew; Fac1: in Taylor1; F2: in out Skew);
  procedure Normalize(F1: in Skew; F2: in out Skew);
  procedure DetFactors(F: in Skew; Fac1,Fac3: in out Taylor1);
  procedure DNormalize(F,DF: in Skew; Fac1,Fac3: in Taylor1; H: in out Skew);

  type PData is -- computed by Prod_GFG for use by DProd_GFG
    record
      R: Radius;
      F,G,G_Right,FG_Right,G_Left: Skew;
    end record;

  procedure Prod_GFG(F,G: in Skew; O: in out PData; H: in out Skew; R: in Radius);
  procedure DProd_GFG(DF,DG: in Skew; O: in PData; DH: in out Skew);

  procedure ShowEigenA10(G: in Skew; UFac,VFac: in Scalar);
  procedure ShowCommutator(F,G: in Skew);

  BigRho: constant Radius := Flt(64);
  SNegOne: constant Scalar := Scal(NegOne);
  STwo: constant Scalar := Scal(Two);
  SHalf: constant Scalar := Scal(Half);
  SNegHalf: constant Scalar := Scal(NegHalf);

private

  --- parallelized products and translations
  procedure Prod3(F1,G1: in Taylor1; H1: in out Taylor1;
                  F2,G2: in Taylor1; H2: in out Taylor1;
                  F3,G3: in Taylor1; H3: in out Taylor1);
  procedure Prod4(F1,G1: in Taylor1; H1: in out Taylor1;
                  F2,G2: in Taylor1; H2: in out Taylor1;
                  F3,G3: in Taylor1; H3: in out Taylor1;
                  F4,G4: in Taylor1; H4: in out Taylor1);
  procedure Prod8(F1,G1: in Taylor1; H1: in out Taylor1;
                  F2,G2: in Taylor1; H2: in out Taylor1;
                  F3,G3: in Taylor1; H3: in out Taylor1;
                  F4,G4: in Taylor1; H4: in out Taylor1;
                  F5,G5: in Taylor1; H5: in out Taylor1;
                  F6,G6: in Taylor1; H6: in out Taylor1;
                  F7,G7: in Taylor1; H7: in out Taylor1;
                  F8,G8: in Taylor1; H8: in out Taylor1);
  procedure Translate4(F1: in Taylor1; S1: in Scalar; G1: in out Taylor1;
                       F2: in Taylor1; S2: in Scalar; G2: in out Taylor1;
                       F3: in Taylor1; S3: in Scalar; G3: in out Taylor1;
                       F4: in Taylor1; S4: in Scalar; G4: in out Taylor1;
                       R: in Radius);

  procedure Inv(M: in out SMat2);
  procedure TUVS2ABCD(M: in out SMat2);
  procedure TUVS2ABCD(F: in out TMat2; Fix_Rev: in Boolean := False);
  procedure TUVS2ABCD(F: in TMat2; G: in out TMat2; Fix_Rev: in Boolean := False);
  procedure ABCD2TUVS(F: in out TMat2; Fix_Rev: in Boolean := False);
  procedure ABCD2TUVS(F: in TMat2; G: in out TMat2; Fix_Rev: in Boolean := False);
  procedure InvConjugate(F: in out TMat2);
  procedure Pre_Prod_GFG(F,G: in Skew; O: in out PData);

  J11: Index4 renames JA;
  J12: Index4 renames JB;
  J21: Index4 renames JC;
  J22: Index4 renames JD;

end Taylors1.Skews;
